@@ -1,11 +1,6 @@
 # -*- coding: utf-8 -*-
 """ This module provides the backend Flask server used by psiTurk. """
-from __future__ import print_function
-from __future__ import absolute_import
-
-from .psiturk_statuses import *
-from builtins import str
-from builtins import range
+from __future__ import generator_stop
 import os
 import sys
 import datetime
@@ -15,37 +10,31 @@ import user_agents
 import requests
 import re
 import json
+from jinja2 import TemplateNotFound
+from collections import Counter
 
-import six
-if six.PY2:
-    # FileNotFoundError is only available since Python 3.3
-    FileNotFoundError = IOError
-    from io import open
-
-try:
-    from collections import Counter
-except ImportError:
-    from .counter import Counter
 
 # Setup flask
 from flask import Flask, render_template, render_template_string, request, \
-    jsonify
+    jsonify, send_from_directory
 
 # Setup database
+
 from .db import db_session, init_db
 from .models import Participant
 from sqlalchemy import or_, exc
 
+from .psiturk_statuses import *
 from .psiturk_config import PsiturkConfig
-from .experiment_errors import ExperimentError, InvalidUsage
-from psiturk.user_utils import nocache
+from .experiment_errors import ExperimentError, ExperimentApiError
+from .user_utils import nocache
 
 # Setup config
 CONFIG = PsiturkConfig()
 CONFIG.load_config()
 
 # Setup logging
-if 'ON_HEROKU' in os.environ:
+if 'ON_CLOUD' in os.environ:
     LOG_FILE_PATH = None
 else:
     LOG_FILE_PATH = os.path.join(os.getcwd(), CONFIG.get("Server Parameters",
@@ -57,16 +46,39 @@ LOG_LEVEL = LOG_LEVELS[CONFIG.getint('Server Parameters', 'loglevel')]
 logging.basicConfig(filename=LOG_FILE_PATH, format='%(asctime)s %(message)s',
                     level=LOG_LEVEL)
 
+
 # Status codes
 
 # Let's start
 # ===========
 
 app = Flask("Experiment_Server")
+
+# experiment server logging
+if 'gunicorn' in os.environ.get('SERVER_SOFTWARE', ''):
+    gunicorn_error_logger = logging.getLogger('gunicorn.error')
+    app.logger.handlers.extend(gunicorn_error_logger.handlers)
+
 # Set cache timeout to 10 seconds for static files
 app.config.update(SEND_FILE_MAX_AGE_DEFAULT=10)
 app.secret_key = CONFIG.get('Server Parameters', 'secret_key')
-app.logger.info("Secret key: " + app.secret_key)
+
+
+def check_templates_exist():
+    # this checks for templates that are required if you are hosting your own ad.
+    try_these = ['thanks-mturksubmit.html', 'closepopup.html']
+    try:
+        try_these = ['thanks-mturksubmit.html', 'closepopup.html']
+        [app.jinja_env.get_template(try_this) for try_this in try_these]
+    except TemplateNotFound as e:
+        raise RuntimeError((
+            f"Missing one of the following templates: {', '.join(try_these)}."
+            f"Copy these over from a freshly-created psiturk example experiment."
+            f"{type(e).__name__, str(e)}"
+        ))
+
+
+check_templates_exist()
 
 
 # Serving warm, fresh, & sweet custom, user-provided routes
@@ -75,20 +87,58 @@ app.logger.info("Secret key: " + app.secret_key)
 try:
     sys.path.append(os.getcwd())
     from custom import custom_code
+except ModuleNotFoundError as e:
+    app.logger.info("Hmm... it seems no custom code (custom.py) associated \
+                    with this project.")
 except ImportError as e:
-    if str(e) == 'No module named custom':
-        app.logger.info("Hmm... it seems no custom code (custom.py) associated \
-                        with this project.")
-    else:
-        app.logger.error("There is custom code (custom.py) associated with this \
-                          project but it doesn't import cleanly.  Raising exception,")
-        raise
+    app.logger.error("There is custom code (custom.py) associated with this \
+                      project but it doesn't import cleanly.  Raising exception,")
+    raise
 else:
     app.register_blueprint(custom_code)
+    try:
+        # noinspection PyUnresolvedReferences
+        from custom import init_app as custom_init_app
+    except ImportError as e:
+        pass
+    else:
+        custom_init_app(app)
+
+# scheduler
+
+from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
+from pytz import utc
+
+from .db import engine
+jobstores = {
+    'default': SQLAlchemyJobStore(engine=engine)
+}
+if 'gunicorn' in os.environ.get('SERVER_SOFTWARE', ''):
+    from apscheduler.schedulers.gevent import GeventScheduler as Scheduler
+else:
+    from apscheduler.schedulers.background import BackgroundScheduler as Scheduler
+logging.getLogger('apscheduler').setLevel(logging.DEBUG)
+scheduler = Scheduler(jobstores=jobstores, timezone=utc)
+app.apscheduler = scheduler
+scheduler.app = app
+
+if CONFIG.getboolean('Server Parameters', 'do_scheduler'):
+    scheduler.start()
+
+
+#
+# Dashboard
+#
+if CONFIG.getboolean('Server Parameters', 'enable_dashboard'):
+    from .dashboard import dashboard, init_app as dashboard_init_app # management dashboard
+    app.register_blueprint(dashboard)
+    dashboard_init_app(app)
+
+    from .api import api_blueprint
+    app.register_blueprint(api_blueprint)
 
 init_db()
 
-
 # Read psiturk.js file into memory
 PSITURK_JS_FILE = os.path.join(os.path.dirname(__file__),
                                "psiturk_js/psiturk.js")
@@ -105,12 +155,13 @@ def handle_exp_error(exception):
     """Handle errors by sending an error page."""
     app.logger.error(
         "%s (%s) %s", exception.value, exception.errornum, str(dict(request.args)))
-    return exception.error_page(request, CONFIG.get('HIT Configuration',
+    return exception.error_page(request, CONFIG.get('Task Parameters',
                                                     'contact_email_on_error'))
 
-# for use with API errors
-@app.errorhandler(InvalidUsage)
-def handle_invalid_usage(error):
+
+@app.errorhandler(ExperimentApiError)
+def handle_experiment_api_error(error):
+    # for use with API errors
     response = jsonify(error.to_dict())
     response.status_code = error.status_code
     app.logger.error(error.message)
@@ -119,7 +170,7 @@ def handle_invalid_usage(error):
 
 @app.teardown_request
 def shutdown_session(_=None):
-    ''' Shut down session route '''
+    """ Shut down session route """
     db_session.remove()
 
 
@@ -138,16 +189,16 @@ def get_random_condcount(mode):
 
     Returns a tuple: (cond, condition)
     """
-    cutofftime = datetime.timedelta(minutes=-CONFIG.getint('Server Parameters',
+    cutofftime = datetime.timedelta(minutes=-CONFIG.getint('Task Parameters',
                                                            'cutoff_time'))
-    starttime = datetime.datetime.now() + cutofftime
+    starttime = datetime.datetime.now(datetime.timezone.utc) + cutofftime
 
     try:
         conditions = json.load(
             open(os.path.join(app.root_path, 'conditions.json')))
         numconds = len(list(conditions.keys()))
         numcounts = 1
-    except IOError as e:
+    except IOError:
         numconds = CONFIG.getint('Task Parameters', 'num_conds')
         numcounts = CONFIG.getint('Task Parameters', 'num_counters')
 
@@ -171,45 +222,49 @@ def get_random_condcount(mode):
     mincount = min(counts.values())
     minima = [hsh for hsh, count in counts.items() if count == mincount]
     chosen = choice(minima)
-    #conds += [ 0 for _ in range(1000) ]
-    #conds += [ 1 for _ in range(1000) ]
     app.logger.info("given %(a)s chose %(b)s" % {'a': counts, 'b': chosen})
 
     return chosen
 
 
+try:
+    from custom import custom_get_condition as get_condition
+except (ModuleNotFoundError, ImportError):
+    get_condition = get_random_condcount
+
 # Routes
 # ======
 
+
 @app.route('/')
 @nocache
 def index():
-    ''' Index route '''
+    """ Index route """
     return render_template('default.html')
 
 
 @app.route('/favicon.ico')
 def favicon():
-    ''' Serve favicon '''
+    """ Serve favicon """
     return app.send_static_file('favicon.ico')
 
 
 @app.route('/static/js/psiturk.js')
 def psiturk_js():
-    ''' psiTurk js route '''
+    """ psiTurk js route """
     return render_template_string(PSITURK_JS_CODE)
 
 
 @app.route('/check_worker_status', methods=['GET'])
 def check_worker_status():
-    ''' Check worker status route '''
+    """ Check worker status route """
     if 'workerId' not in request.args:
         resp = {"status": "bad request"}
         return jsonify(**resp)
     else:
         worker_id = request.args['workerId']
         assignment_id = request.args['assignmentId']
-        allow_repeats = CONFIG.getboolean('HIT Configuration', 'allow_repeats')
+        allow_repeats = CONFIG.getboolean('Task Parameters', 'allow_repeats')
         if allow_repeats:  # if you allow repeats focus on current worker/assignment combo
             try:
                 part = Participant.query.\
@@ -251,7 +306,7 @@ def advertisement():
     user_agent_string = request.user_agent.string
     user_agent_obj = user_agents.parse(user_agent_string)
     browser_ok = True
-    browser_exclude_rule = CONFIG.get('HIT Configuration', 'browser_exclude_rule')
+    browser_exclude_rule = CONFIG.get('Task Parameters', 'browser_exclude_rule')
     for rule in browser_exclude_rule.split(','):
         myrule = rule.strip()
         if myrule in ["mobile", "tablet", "touchcapable", "pc", "bot"]:
@@ -261,7 +316,7 @@ def advertisement():
                (myrule == "pc" and user_agent_obj.is_pc) or\
                (myrule == "bot" and user_agent_obj.is_bot):
                 browser_ok = False
-        elif (myrule == "Safari" or myrule == "safari"):
+        elif myrule == "Safari" or myrule == "safari":
             if "Chrome" in user_agent_string and "Safari" in user_agent_string:
                 pass
             elif "Safari" in user_agent_string:
@@ -305,7 +360,7 @@ def advertisement():
     except exc.SQLAlchemyError:
         status = None
 
-    allow_repeats = CONFIG.getboolean('HIT Configuration', 'allow_repeats')
+    allow_repeats = CONFIG.getboolean('Task Parameters', 'allow_repeats')
     if (status == STARTED or status == QUITEARLY) and not debug_mode:
         # Once participants have finished the instructions, we do not allow
         # them to start the task again.
@@ -314,28 +369,23 @@ def advertisement():
         # 'or status == SUBMITTED' because we suspect that sometimes the post
         # to mturk fails after we've set status to SUBMITTED, so really they
         # have not successfully submitted. This gives another chance for the
-        # submit to work when not using the psiturk ad server.
-        use_psiturk_ad_server = CONFIG.getboolean(
-            'Shell Parameters', 'use_psiturk_ad_server')
-        if not use_psiturk_ad_server:
-            # They've finished the experiment but haven't successfully submitted the HIT
-            # yet.
-            return render_template(
-                'thanks-mturksubmit.html',
-                using_sandbox=(mode == "sandbox"),
-                hitid=hit_id,
-                assignmentid=assignment_id,
-                workerid=worker_id
-            )
-        else:
-            # Show them a thanks message and tell them to go away.
-            return render_template('thanks.html')
+        # submit to work.
+
+        # They've finished the experiment but haven't successfully submitted the HIT
+        # yet.
+        return render_template(
+            'thanks-mturksubmit.html',
+            using_sandbox=(mode == "sandbox"),
+            hitid=hit_id,
+            assignmentid=assignment_id,
+            workerid=worker_id
+        )
     elif already_in_db and not (debug_mode or allow_repeats):
         raise ExperimentError('already_did_exp_hit')
     elif status == ALLOCATED or not status or debug_mode:
         # Participant has not yet agreed to the consent. They might not
         # even have accepted the HIT.
-        with open('templates/ad.html', 'r', encoding='utf-8') as temp_file:
+        with open('templates/ad.html', 'r') as temp_file:
             ad_string = temp_file.read()
         ad_string = insert_mode(ad_string, mode)
         return render_template_string(
@@ -361,7 +411,7 @@ def give_consent():
     assignment_id = request.args['assignmentId']
     worker_id = request.args['workerId']
     mode = request.args['mode']
-    with open('templates/consent.html', 'r', encoding='utf-8') as temp_file:
+    with open('templates/consent.html', 'r') as temp_file:
         consent_string = temp_file.read()
     consent_string = insert_mode(consent_string, mode)
     return render_template_string(
@@ -372,22 +422,6 @@ def give_consent():
     )
 
 
-def get_ad_via_hitid(hit_id):
-    ''' Get ad via HIT id '''
-    username = CONFIG.get('psiTurk Access', 'psiturk_access_key_id')
-    password = CONFIG.get('psiTurk Access', 'psiturk_secret_access_id')
-    try:
-        req = requests.get('https://api.psiturk.org/api/ad/lookup/' + hit_id,
-                           auth=(username, password))
-    except:
-        raise ExperimentError('api_server_not_reachable')
-    else:
-        if req.status_code == 200:
-            return req.json()['ad_id']
-        else:
-            return "error"
-
-
 @app.route('/exp', methods=['GET'])
 @nocache
 def start_exp():
@@ -411,7 +445,7 @@ def start_exp():
 
     # Check first to see if this hitId or assignmentId exists.  If so, check to
     # see if inExp is set
-    allow_repeats = CONFIG.getboolean('HIT Configuration', 'allow_repeats')
+    allow_repeats = CONFIG.getboolean('Task Parameters', 'allow_repeats')
     if allow_repeats:
         matches = Participant.query.\
             filter(Participant.workerid == worker_id).\
@@ -425,7 +459,7 @@ def start_exp():
     numrecs = len(matches)
     if numrecs == 0:
         # Choose condition and counterbalance
-        subj_cond, subj_counter = get_random_condcount(mode)
+        subj_cond, subj_counter = get_condition(mode)
 
         worker_ip = "UNKNOWN" if not request.remote_addr else \
             request.remote_addr
@@ -433,8 +467,8 @@ def start_exp():
             request.user_agent.browser
         platform = "UNKNOWN" if not request.user_agent.platform else \
             request.user_agent.platform
-        language = "UNKNOWN" if not request.user_agent.language else \
-            request.user_agent.language
+        language = "UNKNOWN" if not request.accept_languages else \
+            request.accept_languages.best
 
         # Set condition here and insert into database.
         participant_attributes = dict(
@@ -484,23 +518,7 @@ def start_exp():
             if other_assignment:
                 raise ExperimentError('already_did_exp_hit')
 
-    use_psiturk_ad_server = CONFIG.getboolean(
-        'Shell Parameters', 'use_psiturk_ad_server')
-    if use_psiturk_ad_server and (mode == 'sandbox' or mode == 'live'):
-        # If everything goes ok here relatively safe to assume we can lookup
-        # the ad.
-        ad_id = get_ad_via_hitid(hit_id)
-        if ad_id != "error":
-            if mode == "sandbox":
-                ad_server_location = 'https://sandbox.ad.psiturk.org/complete/'\
-                    + str(ad_id)
-            elif mode == "live":
-                ad_server_location = 'https://ad.psiturk.org/complete/' +\
-                    str(ad_id)
-        else:
-            raise ExperimentError('hit_not_registered_with_ad_server')
-    else:
-        ad_server_location = '/complete'
+    ad_server_location = '/complete'
 
     return render_template(
         'exp.html', uniqueId=part.uniqueid,
@@ -509,7 +527,9 @@ def start_exp():
         adServerLoc=ad_server_location,
         mode=mode,
         contact_address=CONFIG.get(
-            'HIT Configuration', 'contact_email_on_error')
+            'Task Parameters', 'contact_email_on_error'),
+        codeversion=CONFIG.get(
+            'Task Parameters', 'experiment_code_version')
     )
 
 
@@ -523,7 +543,7 @@ def enterexp():
     referesh to start over).
     """
     app.logger.info("Accessing /inexp")
-    if not 'uniqueId' in request.form:
+    if 'uniqueId' not in request.form:
         raise ExperimentError('improper_inputs')
     unique_id = request.form['uniqueId']
 
@@ -531,7 +551,7 @@ def enterexp():
         user = Participant.query.\
             filter(Participant.uniqueid == unique_id).one()
         user.status = STARTED
-        user.beginexp = datetime.datetime.now()
+        user.beginexp = datetime.datetime.now(datetime.timezone.utc)
         db_session.add(user)
         db_session.commit()
         resp = {"status": "success"}
@@ -557,20 +577,19 @@ def load(uid=None):
             one()
     except exc.SQLAlchemyError:
         app.logger.error("DB error: Unique user not found.")
-
-    try:
-        resp = json.loads(user.datastring)
-    except:
-        resp = {
-            "condition": user.cond,
-            "counterbalance": user.counterbalance,
-            "assignmentId": user.assignmentid,
-            "workerId": user.workerid,
-            "hitId": user.hitid,
-            "bonus": user.bonus
-        }
-
-    return jsonify(**resp)
+    else:
+        try:
+            resp = json.loads(user.datastring)
+        except (ValueError, TypeError, json.JSONDecodeError):
+            resp = {
+                "condition": user.cond,
+                "counterbalance": user.counterbalance,
+                "assignmentId": user.assignmentid,
+                "workerId": user.workerid,
+                "hitId": user.hitid,
+                "bonus": user.bonus
+            }
+        return jsonify(**resp)
 
 
 @app.route('/sync/<uid>', methods=['PUT'])
@@ -586,16 +605,16 @@ def update(uid=None):
             filter(Participant.uniqueid == uid).\
             one()
     except exc.SQLAlchemyError:
-        app.logger.error("DB error: Unique user not found.")
+        raise ExperimentApiError("DB error: Unique user not found.")
 
-    if hasattr(request, 'json'):
-        user.datastring = json.dumps(request.json)
-        db_session.add(user)
-        db_session.commit()
+    user.datastring = json.dumps(request.json)
+    db_session.add(user)
+    db_session.commit()
 
-    
-    data = json.loads(user.datastring)
-    
+    try:
+        data = json.loads(user.datastring)
+    except Exception as e:
+        raise ExperimentApiError('failed to load json datastring back from database as object! Error was {}: {}'.format(type(e), str(e)))
 
     trial = data.get("currenttrial", None)
     app.logger.info("saved data for %s (current trial: %s)", uid, trial)
@@ -638,8 +657,8 @@ def quitter():
 @app.route('/complete', methods=['GET'])
 @nocache
 def debug_complete():
-    ''' Debugging route for complete. '''
-    if not 'uniqueId' in request.args:
+    """ Debugging route for complete. """
+    if 'uniqueId' not in request.args:
         raise ExperimentError('improper_inputs')
     else:
         unique_id = request.args['uniqueId']
@@ -648,23 +667,26 @@ def debug_complete():
             user = Participant.query.\
                 filter(Participant.uniqueid == unique_id).one()
             user.status = COMPLETED
-            user.endhit = datetime.datetime.now()
+            user.endhit = datetime.datetime.now(datetime.timezone.utc)
             db_session.add(user)
             db_session.commit()
-        except:
+        except exc.SQLAlchemyError:
             raise ExperimentError('error_setting_worker_complete')
         else:
             # send them back to mturk.
-            if (mode == 'sandbox' or mode == 'live'):
+            if mode == 'sandbox' or mode == 'live':
                 return render_template('closepopup.html')
             else:
-                return render_template('complete.html')
+                allow_repeats = CONFIG.getboolean('Task Parameters', 'allow_repeats')
+                return render_template('complete.html',
+                                       allow_repeats=allow_repeats,
+                                       worker_id=user.workerid)
 
 
 @app.route('/worker_complete', methods=['GET'])
 def worker_complete():
-    ''' Complete worker. '''
-    if not 'uniqueId' in request.args:
+    """ Complete worker. """
+    if 'uniqueId' not in request.args:
         resp = {"status": "bad request"}
         return jsonify(**resp)
     else:
@@ -674,7 +696,7 @@ def worker_complete():
             user = Participant.query.\
                 filter(Participant.uniqueid == unique_id).one()
             user.status = COMPLETED
-            user.endhit = datetime.datetime.now()
+            user.endhit = datetime.datetime.now(datetime.timezone.utc)
             db_session.add(user)
             db_session.commit()
             status = "success"
@@ -686,8 +708,8 @@ def worker_complete():
 
 @app.route('/worker_submitted', methods=['GET'])
 def worker_submitted():
-    ''' Submit worker '''
-    if not 'uniqueId' in request.args:
+    """ Submit worker """
+    if 'uniqueId' not in request.args:
         resp = {"status": "bad request"}
         return jsonify(**resp)
     else:
@@ -708,7 +730,7 @@ def worker_submitted():
 # Is this a security risk?
 @app.route("/ppid")
 def ppid():
-    ''' Get ppid '''
+    """ Get ppid """
     proc_id = os.getppid()
     return str(proc_id)
 
@@ -717,7 +739,7 @@ def ppid():
 
 
 def insert_mode(page_html, mode):
-    ''' Insert mode '''
+    """ Insert mode """
     page_html = page_html
     match_found = False
     matches = re.finditer('workerId={{ workerid }}', page_html)
@@ -735,25 +757,19 @@ def insert_mode(page_html, mode):
 # Generic route
 # =============
 
-@app.route('/<pagename>')
-@app.route('/<foldername>/<pagename>')
-def regularpage(foldername=None, pagename=None):
+@app.route('/<path:path>')
+def regularpage(path):
     """
     Route not found by the other routes above. May point to a static template.
     """
-    if foldername is None and pagename is None:
-        raise ExperimentError('page_not_found')
-    if foldername is None and pagename is not None:
-        return render_template(pagename)
-    else:
-        return render_template(foldername+"/"+pagename)
+    return send_from_directory('templates', path)
 
 
 def run_webserver():
-    ''' Run web server '''
-    host = "0.0.0.0"
+    """ Run web server """
+    host = CONFIG.get('Server Parameters', 'host')
     port = CONFIG.getint('Server Parameters', 'port')
-    print("Serving on ", "http://" + host + ":" + str(port))
+    print(f"Serving on http://{host}:{port}")
     app.config['TEMPLATES_AUTO_RELOAD'] = True
     app.jinja_env.auto_reload = True
     app.run(debug=True, host=host, port=port)
